//
//  MBCoreDataTableViewController.m
//
//  Created by Martina Brazelton on 6/1/13.
//  Copyright (c) 2013 Martina Brazelton. All rights reserved.
//

#import "MBCoreDataTableViewController.h"

NSString *const CELL_REUSE_ID = @"CellIdentifier";
NSString *const HEADER_REUSE_ID = @"HeaderIdentifier";

@interface MBCoreDataTableViewController ()

@property (nonatomic, strong) NSFetchedResultsController *fetchedResultsController;

@end

@implementation MBCoreDataTableViewController

- (id)initWithFetchedResultsController:(NSFetchedResultsController *)fetchedResultsController
{
    self = [super initWithStyle:UITableViewStylePlain];
    if (self) {
        self.fetchedResultsController = fetchedResultsController;
        self.context = self.fetchedResultsController.managedObjectContext;
    }
    
    return self;
}

+ (MBCoreDataTableViewController *)coreDataTableViewControllerWithFetchedResultsController:(NSFetchedResultsController *)fetchedResultsController
{
    MBCoreDataTableViewController *coreDataTableViewController = [MBCoreDataTableViewController.alloc initWithFetchedResultsController:fetchedResultsController];
    
    return coreDataTableViewController;
}

- (Class)cellClass
{
    return UITableViewCell.class;
}

- (Class)headerClass
{
    return UITableViewHeaderFooterView.class;
}

- (NSString *)cellReuseID
{
    return CELL_REUSE_ID;
}

- (NSString *)headerReuseID
{
    return HEADER_REUSE_ID;
}

+ (NSFetchedResultsController *)fetchedResultsControllerWithFetchRequest:(NSFetchRequest *)fetchRequest context:(NSManagedObjectContext *)context sectionNameKeyPath:(NSString *)sectionNameKeyPath cacheName:(NSString *)cacheName delegate:(id<NSFetchedResultsControllerDelegate>)delegate
{
    NSFetchedResultsController *frc = [NSFetchedResultsController.alloc
                                             initWithFetchRequest:fetchRequest
                                             managedObjectContext:context
                                               sectionNameKeyPath:sectionNameKeyPath
                                                        cacheName:cacheName];
    frc.delegate = delegate;
    
    return frc;
}

+ (NSFetchRequest *)fetchRequestWithEntity:(NSString *)entityName sortDescriptors:(NSArray *)sortDescriptors predicate:(NSPredicate *)predicate
{
    NSFetchRequest *request = [NSFetchRequest.alloc initWithEntityName:entityName];
    request.sortDescriptors = sortDescriptors;
    if (predicate)
        request.predicate = predicate;
    
    return request;
}

- (BOOL)performFetch
{
    if (!self.fetchedResultsController) {
        return NO;  
    }
    
    NSError *error = nil;
    BOOL success = [self.fetchedResultsController performFetch:&error];
    if (error) {
        [self handleFetchError:error];
    }
    [self.tableView reloadData];
    
    return success;
}

- (void)handleFetchError:(NSError *)error
{
    NSLog(@"Error performing fetch: %@", error.description);   
}

- (void)handleSaveError:(NSError *)error
{
    NSLog(@"Error performing save: %@", error.description);
}

- (BOOL)saveChanges
{
    BOOL saved = NO;
    
    if (self.context.hasChanges) {
        NSError *error;
        saved = [self.context save:&error];
        if (error) {
            [self handleSaveError:error];
        }
    }
    
    return saved;
}

- (id <NSFetchedResultsSectionInfo>)sectionAtIndex:(NSInteger)section
{
    return [self.fetchedResultsController.sections objectAtIndex:section];
}

- (NSManagedObject *)managedObjectAtIndexPath:(NSIndexPath *)indexPath
{
    return [self.fetchedResultsController objectAtIndexPath:indexPath];
}

- (NSIndexPath *)indexPathForManagedObject:(NSManagedObject *)managedObject
{
    return [self.fetchedResultsController indexPathForObject:managedObject];
}

- (NSManagedObject *)managedObjectAtCell:(UITableViewCell *)cell
{
    NSIndexPath *indexPath = [self.tableView indexPathForCell:cell];
    return [self managedObjectAtIndexPath:indexPath];
}

- (UITableViewCell *)cellForManagedObject:(NSManagedObject *)managedObject
{
    return [self.tableView cellForRowAtIndexPath:[self indexPathForManagedObject:managedObject]];
}

- (void)configureCell:(UITableViewCell *)cell forIndexPath:(NSIndexPath *)indexPath
{
    
}

- (void)configureHeader:(UIView *)header forSection:(NSInteger)section
{
    
}

- (void)didSelectManagedObject:(NSManagedObject *)managedObject
{
    
}

#pragma mark - NSFetchedResultsController delegate

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type newIndexPath:(NSIndexPath *)newIndexPath
{
    switch (type) {
        case NSFetchedResultsChangeDelete:
            [self.tableView deleteRowsAtIndexPaths:@[indexPath] withRowAnimation:UITableViewRowAnimationFade];
            break;
        case NSFetchedResultsChangeInsert:
            [self.tableView insertRowsAtIndexPaths:@[indexPath] withRowAnimation:UITableViewRowAnimationFade];
            break;
            
        case NSFetchedResultsChangeUpdate:
            [self.tableView reloadRowsAtIndexPaths:@[indexPath] withRowAnimation:UITableViewRowAnimationFade];
            break;
        default:
            break;
    }
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    return self.fetchedResultsController.sections.count;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    id <NSFetchedResultsSectionInfo> sectionInfo = [self sectionAtIndex:section];
    
    return sectionInfo.numberOfObjects;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:self.cellReuseID forIndexPath:indexPath];
    
    [self configureCell:cell forIndexPath:indexPath];
    
    return cell;
}

- (CGFloat)tableView:(UITableView *)tableView heightForHeaderInSection:(NSInteger)section
{
    return 40;
}

- (UIView *)tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
{
    UIView *header = [self.tableView dequeueReusableHeaderFooterViewWithIdentifier:self.headerReuseID];
    [self configureHeader:header forSection:section];
    
    return header;
}

#pragma mark - Table view delegate

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    [self didSelectManagedObject:[self managedObjectAtIndexPath:indexPath]];
}

#pragma mark - View Lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    [self.tableView registerClass:self.cellClass forCellReuseIdentifier:self.cellReuseID];
    [self.tableView registerClass:self.headerClass forHeaderFooterViewReuseIdentifier:self.headerReuseID];
}

- (void)viewDidAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    [self performFetch];
}

@end
