//
//  MBReorderableTableView.m
//  ListedApp
//
//  Created by Martina Brazelton on 6/8/13.
//  Copyright (c) 2013 Martina Brazelton. All rights reserved.
//

#import "MBReorderableTableView.h"
#import <QuartzCore/QuartzCore.h>

@interface MBReorderableTableView ()

@property (nonatomic, strong) UILongPressGestureRecognizer *longPress;
@property (nonatomic, strong) NSTimer *scrollTimer;
@property (nonatomic, assign) CGFloat scrollRate;
@property (nonatomic, strong) NSIndexPath *oldIndexPath;
@property (nonatomic, strong) NSIndexPath *currentIndexPath;
@property (nonatomic, strong) UIImageView *cellSnapshot;
@property (nonatomic, strong) id cellObject;

- (void)initialize;
- (void)setCanReorder:(BOOL)canReorder;
- (void)handleLongPress:(UILongPressGestureRecognizer *)gesture;
- (void)prepareToReorder:(UILongPressGestureRecognizer *)gesture;
- (void)reorderInProgressAtLocation:(CGPoint)location;
- (void)finishReordering;
- (void)startReorder;
- (void)performReorder:(NSIndexPath *)indexPath;
- (void)doneReordering;
- (BOOL)shouldReorder:(UILongPressGestureRecognizer *)gesture;
- (CGFloat)rowsCount;
- (void)deselectCell:(UITableViewCell *)cell;
- (UIImage *)snapshotImageForCell:(UITableViewCell *)cell;
- (UIImageView *)snapshotImageViewForCellAtIndexPath:(NSIndexPath *)indexPath;
- (void)addDropShadowToSnapshot:(UIImageView *)snapshot;
- (void)zoomSnapshot:(UIImageView *)snapshot atLocation:(CGPoint)location;
- (void)centerSnapshotToLocation:(CGPoint)location;
- (CGPoint)centerForLocation:(CGPoint)location;
- (NSIndexPath *)indexPathOfLongPress:(UILongPressGestureRecognizer *)gesture;
- (void)beginUpdatesWithRowAnimation:(UITableViewRowAnimation)rowAnimation indexPathToDelete:(NSIndexPath *)deleteIndexPath indexPathToInsert:(NSIndexPath *)insertIndexPath;
- (void)reloadVisibleRows;
- (void)reorderCellsAtLocation:(CGPoint)location;
- (void)scheduleScrollForGesture:(UILongPressGestureRecognizer *)gesture;
- (void)scrollWithTimer:(NSTimer *)timer;
- (CGFloat)scrollRateForGestureLocation:(CGPoint)location;
- (CGPoint)offsetWithScrollRate:(CGFloat)scrollRate;
- (void)reset;
- (void)invalidateScrollTimer;

@end

@implementation MBReorderableTableView

@dynamic canReorder;

- (id)init
{
    self = [super init];
    if (self) {
        [self initialize];
    }
    
    return self;
}

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self initialize];
    }
    
    return self;
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
    self = [super initWithCoder:aDecoder];
    if (self) {
        [self initialize];
    }
    
    return self;
}

- (id)initWithFrame:(CGRect)frame style:(UITableViewStyle)style
{
    self = [super initWithFrame:frame style:style];
    if (self) {
        [self initialize];
    }
    
    return self;
}

- (void)initialize
{
    self.longPress = [[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(handleLongPress:)];
    [self addGestureRecognizer:self.longPress];
    
    self.canReorder = YES;
}

- (void)setCanReorder:(BOOL)canReorder
{
    self.longPress.enabled = canReorder;
}

- (void)handleLongPress:(UILongPressGestureRecognizer *)gesture
{
    if (![self shouldReorder:gesture]) {
        return;
    }
    
    if (gesture.state == UIGestureRecognizerStateBegan) {
        [self prepareToReorder:gesture];
    } else if (gesture.state == UIGestureRecognizerStateChanged) {
        CGPoint location = [gesture locationInView:self];
        [self reorderInProgressAtLocation:location];
    } else if (gesture.state == UIGestureRecognizerStateEnded) {
        [self finishReordering];
    }
}

- (void)startReorder
{
    NSIndexPath *indexPath = self.oldIndexPath;
    [self beginUpdatesWithRowAnimation:UITableViewRowAnimationNone indexPathToDelete:indexPath indexPathToInsert:indexPath];
    self.cellObject = self.startReorderHandler(self, indexPath);
    self.currentIndexPath = indexPath;
    [self endUpdates];
}

- (void)performReorder:(NSIndexPath *)indexPath
{
    [self beginUpdatesWithRowAnimation:UITableViewRowAnimationAutomatic indexPathToDelete:self.currentIndexPath indexPathToInsert:indexPath];
    self.reorderHandler(self, self.currentIndexPath, indexPath);
    self.currentIndexPath = indexPath;
    [self endUpdates];
}

- (void)doneReordering
{
    NSIndexPath *indexPath = self.currentIndexPath;
    [self beginUpdatesWithRowAnimation:UITableViewRowAnimationNone indexPathToDelete:indexPath indexPathToInsert:indexPath];
    self.completionHandler(self, self.oldIndexPath, indexPath, self.cellObject);
    [self endUpdates];
}

- (void)prepareToReorder:(UILongPressGestureRecognizer *)gesture
{
    CGPoint location = [gesture locationInView:self];
    NSIndexPath *indexPath = [self indexPathForRowAtPoint:location];
    UITableViewCell *cell = [self cellForRowAtIndexPath:indexPath];
    [self deselectCell:cell];
    
    self.oldIndexPath = indexPath;
    self.currentIndexPath = indexPath;
    
    if (!self.cellSnapshot) {
        self.cellSnapshot = [self snapshotImageViewForCellAtIndexPath:indexPath];
        [self addSubview:self.cellSnapshot];
        [self addDropShadowToSnapshot:self.cellSnapshot];
        [self zoomSnapshot:self.cellSnapshot atLocation:location];
    }
    
    [self startReorder];
    
    [self scheduleScrollForGesture:gesture];
}

- (void)reorderInProgressAtLocation:(CGPoint)location
{
    [self centerSnapshotToLocation:location];
    
    [self reorderCellsAtLocation:location];
    
    self.scrollRate = [self scrollRateForGestureLocation:location];
}

- (void)reorderCellsAtLocation:(CGPoint)location
{
    NSIndexPath *indexPath = [self indexPathForRowAtPoint:location];
    
    if (indexPath && ![indexPath isEqual:self.currentIndexPath]) {
        [self performReorder:indexPath];
    }
}

- (void)finishReordering
{    
    [UIView animateWithDuration:ANIMATION_DURATION
                     animations:^{
                         CGRect rect = [self rectForRowAtIndexPath:self.currentIndexPath];
                         self.cellSnapshot.transform = CGAffineTransformIdentity;
                         self.cellSnapshot.frame = CGRectOffset(self.cellSnapshot.bounds, rect.origin.x, rect.origin.y);
                     } completion:^(BOOL finished) {
                         [self.cellSnapshot removeFromSuperview];
                         
                         [self doneReordering];
                         
                         [self reloadVisibleRows];
                         
                         [self reset];
                     }];
}

- (BOOL)shouldReorder:(UILongPressGestureRecognizer *)gesture
{
    NSIndexPath *indexPath = [self indexPathOfLongPress:gesture];
    
    return !(self.rowsCount == 0 || (gesture.state == UIGestureRecognizerStateBegan && indexPath == nil) ||
             (gesture.state == UIGestureRecognizerStateEnded && self.currentIndexPath == nil));
}

- (CGFloat)rowsCount
{
    int rowsCount = 0;
    for (int i= 0; i < self.numberOfSections; i++) {
        rowsCount += [self numberOfRowsInSection:i];
    }
    
    return rowsCount;
}

- (void)deselectCell:(UITableViewCell *)cell
{
    [cell setSelected:NO animated:NO];
    [cell setHighlighted:NO animated:NO];
}

- (UIImage *)snapshotImageForCell:(UITableViewCell *)cell
{
    UIGraphicsBeginImageContextWithOptions(cell.bounds.size, NO, 0);
    [cell.layer renderInContext:UIGraphicsGetCurrentContext()];
    UIImage *cellImage = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();
    
    return cellImage;
}

- (UIImageView *)snapshotImageViewForCellAtIndexPath:(NSIndexPath *)indexPath
{
    UITableViewCell *cell = [self cellForRowAtIndexPath:indexPath];
    UIImage *cellImage = [self snapshotImageForCell:cell];
    
    UIImageView *snapshot = [UIImageView.alloc initWithImage:cellImage];
    CGRect rect = [self rectForRowAtIndexPath:indexPath];
    snapshot.frame = CGRectOffset(snapshot.bounds, rect.origin.x, rect.origin.y);
    
    return snapshot;
}

- (void)addDropShadowToSnapshot:(UIImageView *)snapshot
{
    snapshot.layer.masksToBounds = NO;
    snapshot.layer.shadowColor = [[UIColor blackColor] CGColor];
    snapshot.layer.shadowOffset = CGSizeMake(0, 0);
    snapshot.layer.shadowRadius = SHADOW_RADIUS;
    snapshot.layer.shadowOpacity = SHADOW_OPACITY;
}

- (void)zoomSnapshot:(UIImageView *)snapshot atLocation:(CGPoint)location
{
    [UIView beginAnimations:@"zoom" context:nil];
    snapshot.transform = CGAffineTransformMakeScale(ZOOM_SCALE, ZOOM_SCALE);
    snapshot.center = [self centerForLocation:location];
    [UIView commitAnimations];
}

- (void)centerSnapshotToLocation:(CGPoint)location
{
    if (location.y >= 0 && location.y <= self.contentSize.height) {
        self.cellSnapshot.center = [self centerForLocation:location];
    }
}

- (CGPoint)centerForLocation:(CGPoint)location
{
    return CGPointMake(self.center.x, location.y);
}

- (NSIndexPath *)indexPathOfLongPress:(UILongPressGestureRecognizer *)gesture
{
    CGPoint location = [gesture locationInView:self];
    return [self indexPathForRowAtPoint:location];
}

- (void)beginUpdatesWithRowAnimation:(UITableViewRowAnimation)rowAnimation indexPathToDelete:(NSIndexPath *)deleteIndexPath indexPathToInsert:(NSIndexPath *)insertIndexPath
{
    [self beginUpdates];
    [self deleteRowsAtIndexPaths:@[ deleteIndexPath ] withRowAnimation:rowAnimation];
    [self insertRowsAtIndexPaths:@[ insertIndexPath ] withRowAnimation:rowAnimation];
}

- (void)reloadVisibleRows
{
    NSMutableArray *visibleRows = self.indexPathsForVisibleRows.mutableCopy;
    [visibleRows removeObject:self.currentIndexPath];
    [self reloadRowsAtIndexPaths:visibleRows withRowAnimation:UITableViewRowAnimationNone];
}

- (void)scheduleScrollForGesture:(UILongPressGestureRecognizer *)gesture
{
    NSMutableDictionary *userInfo = [NSMutableDictionary dictionaryWithObject:gesture forKey:@"gesture"];
    self.scrollTimer = [NSTimer timerWithTimeInterval:SCROLL_TIMER_INTERVAL target:self selector:@selector(scrollWithTimer:) userInfo:userInfo repeats:YES];
    [[NSRunLoop mainRunLoop] addTimer:self.scrollTimer forMode:NSDefaultRunLoopMode];
}

- (void)scrollWithTimer:(NSTimer *)timer
{
    UILongPressGestureRecognizer *gesture = [timer.userInfo objectForKey:@"gesture"];
    CGPoint location  = [gesture locationInView:self];
    
    CGPoint offset = [self offsetWithScrollRate:self.scrollRate];
    [self setContentOffset:offset];

    [self centerSnapshotToLocation:location];
    
    [self reorderCellsAtLocation:location];
}

- (CGFloat)scrollRateForGestureLocation:(CGPoint)location
{
    CGFloat scrollRate = 0;
    CGRect rect = self.bounds;
    CGFloat scrollZoneHeight = rect.size.height / TABLE_VIEW_ZONES;
    CGFloat bottomScrollZone = self.contentOffset.y + rect.size.height - scrollZoneHeight;
    CGFloat topScrollZone = self.contentOffset.y + scrollZoneHeight;
    
    if (location.y >= bottomScrollZone) {
        scrollRate = (location.y - bottomScrollZone) / scrollZoneHeight;
    } else if (location.y <= topScrollZone) {
        scrollRate = (location.y - topScrollZone) / scrollZoneHeight;
    }
    
    return scrollRate;
}

- (CGPoint)offsetWithScrollRate:(CGFloat)scrollRate
{
    CGPoint currentOffset = self.contentOffset;
    CGPoint newOffset = CGPointMake(currentOffset.x, currentOffset.y + scrollRate);
    if (newOffset.y < 0) {
        newOffset.y = 0;
    } else if (self.contentSize.height < self.frame.size.height) {
        newOffset = currentOffset;
    } else if (newOffset.y > self.contentSize.height - self.frame.size.height) {
        newOffset.y = self.contentSize.height - self.frame.size.height;
    }
    
    return newOffset;
}

- (void)reset
{
    [self invalidateScrollTimer];
    self.scrollRate = 0;
    self.oldIndexPath = nil;
    self.currentIndexPath = nil;
    self.cellSnapshot = nil;
}

- (void)invalidateScrollTimer
{
    [self.scrollTimer invalidate];
    self.scrollTimer = nil;
}

- (void)dealloc
{
    [self invalidateScrollTimer];
}

@end
